;;;   literate-misc.el
;;;  this file is under construction.

;;;   Title:  literate header
;;;
;;;   Description:  The following function produces a fortran heading at the 
;;;                 top of the current buffer. It is based on ada-header found
;;;                 in the standard emacs distribution. It can also add in the
;;;                 RCS (revision control system) and last edit date strings
;;;                 tags with a prefix argument. Will attempt to fill the
;;;                 description using latex-auto-fill. 
;;;                 
;;;                 Also included at bottom is blink-or-goto-paren-match which
;;;                 will move the cursor backwards or forwards in the current
;;;                 buffer to a matching parenthesis of the type (,[,or {.
;;;                 Useful for complicated fortran expressions. Both bound to 
;;;                 C-c { and C-c }.
;;;
;;;   Created:  Thu May 28 10:59:30 1992
;;;   Author:   Lawrence R. Dodd
;;;             <dodd@roebling.poly.edu>
;;;		Hacked by Adriaan Joubert from fortran-header
;;;   

;;;   INSTALLATION: 
;;; 
;;;     o  save as literate-header.el in the GNU emacs load-path.
;;;  
;;;     o  stick (autoload 'literate-header "literate-header"
;;;   		     "Code for producing a comment header in Literate routines."
;;;    		     t)
;;;        inside your .emacs, preferably inside literate-mode-hook:
;;;
;;;        (setq literate-mode-hook 
;;;           '(lambda () 
;;;                    .
;;;                    . other stuff
;;;                    .
;;;              (define-key literate-mode-map "\C-ch" 'literate-header)
;;;
;;;              (autoload 'literate-header "literate-header"
;;;                  "Code for producing a comment header in Literate 
;;;                  routines." t)))

;;;   USAGE INSTRUCTIONS: 
;;; 
;;;     o type M-x literate-header or simply C-c h
;;; 
;;;     o type M-1 M-x literate-header or simply M-1 C-c h to get string tags:
;;;        oo will insert standard RCS version string tags (RCS can be 
;;;           obtained via anonymous ftp at 
;;;           /prep.ai.mit.edu:/pub/gnu/rcs-*.tar.Z).
;;;        oo will insert standard LastEditDate string tags (written by Randal 
;;;           L. Schwartz (merlyn@iWarp.intel.com)).
;;;        oo if the variable literate-header-rcs is set to true this is done 
;;;           by default.

;;;   KEY BINDINGS:

;;;     o literate-header is bound to C-c h
;;;     o blink-or-goto-paren-match is bound to C-c { and C-c }
;;;       (I bind it globally to M-{ and M-} as well).

;;;   SEE ALSO:
;;;
;;;     /hallc1.cebaf.gov:/emacs for the most recent version of fortran-mode
;;;     and also for Ralph Finch's fortran-beautifier stored in
;;;     fortran-misc.el
;;;
;;;     /archive.cis.ohio-state.edu:/pub/gnu/emacs/elisp-archive/misc for
;;;     fort-clean.el code for cleaning and stylizing FORTRAN code

;;;     LCD Archive Entry:
;;;     literate-header|Lawrence R. Dodd|dodd@roebling.poly.edu
;;;     |code for producing a comment header in FORTRAN routines
;;;     |$Date: 1992/11/14 19:09:04 $|$Revision: 1.29 $|

(defun lpac-header ()
  "Inserts a header for LPAC literate files"
  (interactive)
  (save-excursion
    (goto-char (point-min)) ; go to top of buffer
    (open-line 1)           ; open a line
    
    ;; read in title string
    (insert "% RCS Information\n")
    (insert "%\n% Title: ")
    (insert (read-string "Brief Title: "))
    
    ;; read in (possibly long) description string
    (insert "\n%\n% Description:  ")
    (insert (read-string "Description (use C-x ^ if needed): "))

    ;; now fill description
    (setq desc-end (point-marker))
    (beginning-of-line)

    ;; fill description
    (setq fill-prefix "%              ")
    (fill-region (point-marker) desc-end t)
    (setq fill-prefix "")
    
    ;; insert LPAC copyright string
    (insert "%\n% Copyright (c) London Parallel Applications Centre\n%\n")

    ;; write creation time and author
    (insert "% $" "Revision$\n")
    (insert "% $" "Date$\n")
    (insert "% $" "Author$\n")
    (insert "% $" "Source$\n%\n")
    (insert "% $" "Log$\n%\n")
    
    ;; Insert revision number and date
    (insert "\\renewcommand{\\RCSrev}{$Revision$}\n")
    (insert "\\renewcommand{\\RCSdate}{$Date$}\n")

    ;; remove tabs and extraneous blank lines
    (untabify (point-min) (point))
    (delete-blank-lines)
    
    ))


(defun include-file ()
  "call find-file with the word under the current point position 
beta (literate only) version"
  (interactive)
  (find-file (concat (expand-file-name (get-string-under-pointer)) '".lit")))


(defun get-string-under-pointer ()
  "return the word around the pointer - this is based on find-tag-default
well alright it is find-tag-default this works ok for tex and literate files
but not c includes change when rest of include file is set up"
  (interactive)
  (save-excursion
    (while (looking-at "\\sw\\|\\s_")
      (forward-char 1))
    (if (re-search-backward "\\sw\\|\\s_" nil t)
	(progn (forward-char 1)
	       (buffer-substring (point)
				 (progn (forward-sexp -1)
					(while (looking-at "\\s'")
					  (forward-char 1))
					(point))))
      nil)))
