/*============================================================================
   PROGRAM      MERGE SORT
   Written by   D. Kim   (dkim@home.korea.ac.kr)
   Date         July 30, 1996
   Description:

This is a merge sort program using BSPlib of Oxford Parallel.
The number of data to be sorted is given by SIZE, and the number of processors
used is decalared by NPROC (currently #PROC=4, #data=32).
The data array is splited into the number of processors with equal
number of data.
The partitioned array each is sorted sequentially by each assigned
processor
and then merged in log (#PROCS) steps.
Superstep synchronization is made at the merging process.

============================================================================*/

#include "bsp.h"
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#define SIZE 32
#define NPROC 4


void bsp_msort(int *my_array, int *your_array, int n, int *out_array) {
  int i;
  int my_ptr, your_ptr;

  for (i=0, my_ptr=0, your_ptr = 0; i<2*n ; i++) {
    if ((my_ptr <n)&&(your_ptr<n))
       if (my_array[my_ptr] < your_array[your_ptr])
           out_array[i] = my_array[my_ptr++];
        else 
           out_array[i] = your_array[your_ptr++];
     else if (my_ptr ==n) 
              out_array[i] = your_array[your_ptr++];
            else if (your_ptr ==n)
              out_array[i] = my_array[my_ptr++];
                 else printf("Something wrong in comparison\n "); 
  }
}

void bsp_exchg_sort(int *array, int curr_size) {
  int i, curr_bound, j, jj;
  int *result= calloc(curr_size,sizeof(int));
  int *your_data = calloc(curr_size,sizeof(int));
  int *mine = calloc(curr_size,sizeof(int));
  bsp_pushregister(your_data,curr_size*sizeof(int));
  bsp_sync();

  curr_bound = curr_size/NPROC;
  
  for (j=0; j< curr_size; j++)
  if (bsp_pid() == j /(curr_size/NPROC))  
    mine[j%(curr_size/NPROC)] = array[j];

  for(i=1;i<NPROC ;i*=2) {

    for (j=0;j < NPROC; j+=2*i){
          if ((bsp_pid()- i ) % (2*i) == 0){

              bsp_put(bsp_pid() -i, mine, your_data,0,curr_bound*sizeof(int));
          }
         bsp_sync();

         for (j=0;j < NPROC; j+=2*i)
         if (bsp_pid() == j) bsp_msort(mine, your_data, curr_bound,result);
        bsp_sync();


    }
    for (j=0;j < NPROC; j+=2*i)
         if (bsp_pid() == j)    
               for (jj=0; jj < 2* curr_bound; jj++)
                  mine[jj] = result[jj];
    curr_bound *= 2;
    bsp_sync();

   }
  bsp_popregister(your_data);
/* NOW COPY THE SORTED ARRAY TO MAIN ROUTINE */
         if (bsp_pid() == 0)    
               for (jj=0; jj < curr_size; jj++)
                  array[jj] = result[jj];
}

void print_array(int *array, int n){
  int i,j;
    for (i=0; i < NPROC; i++){
     if ( i== bsp_pid()){ 
     printf("\n INIT. DATA AT PID_%d = \n", bsp_pid());
      for(j=0;j<n;j++) printf("%d ",array[j]);
      printf("\n");
      fflush(stdout);
   }
  }
}
void init_sort(int *dat,int m, int pid){
  int i,j,amin, index,tmp;

  for (i=0; i< m; i++) {
     amin = dat[i];
     tmp =dat[i];
     index =i;
     for (j=i; j< m; j++)
       if (dat[j] < amin){
          index=j;
          amin = dat[j];
         }
     dat[i] = amin;
     dat[index] = tmp;
   }
}

void copy_array(int *array, int *result, int n, int pid){
     int i;
     int  start = pid * (SIZE/NPROC);
     for  (i=0; i< n/NPROC; i++)
       array[start+i] = result[i];
   }

void init(int *array, int n){
  int start;
  int i,j;
  int *result= calloc(n, sizeof(int));
  int *my_array= calloc(SIZE/NPROC, sizeof(int));
  for (i=0; i< NPROC; i++)
  if (bsp_pid()==i) {
      start = i * (SIZE/NPROC);
     for (j=0; j<SIZE/NPROC ; j++)
      my_array[j] = array[start +j];
     init_sort(my_array, SIZE/NPROC,i);
     copy_array(array, my_array,n,i); 
  }
}

void main() {
  int i,j,n,*xs;
  bsp_begin(bsp_nprocs());
  n=SIZE;
  xs = calloc(n,sizeof(int));

  for (i=0;i<n;i++) 
    if ( i%2 ) xs[i]= i % n;
      else xs[i] = (n-i) % n;
  init(xs,n); /* Sequentially sort and placement to each proc before merging */
  if (bsp_pid()==0)  print_array(xs,n);
  bsp_exchg_sort(xs,n); 

    if (bsp_pid()==0) {
      printf("  FINAL SORTED ARRAY at PROC(%d)\n ",bsp_pid());
      for(j=0;j<n;j++) printf("%d ",xs[j]);
      printf("\n");
      fflush(stdout);
    }
/*    bsp_sync(); */
  
  bsp_end();
}

