/*============================================================================
   PROGRAM      QUICK SORT
   Written by   D. Kim   (dkim@home.korea.ac.kr)
   Date         July 30, 1996
   Description:

This is a quick sort program using BSPlib of Oxford Parallel.
The number of data to be sorted is given by SIZE, and the number of processors
used is decalared by NPROC (currently #PROCS=8, #data=32).
The data array is splited into two parts, left_array and right_array,
and the left_array is given to the spowning processor, and the right_array
to the processor with the PDI= bsp_pid() + NPROC/(int) pow(2,level) 
if the latter is an integer, otherwise the array is sequentially sorted 
by the processor.
Level starts from 1 at the beginning of the sort, then increments 1 at each
new split of the array.
The sort terminates when all numbers are sorted, which is judged by the count 
of sorted number stored at count_buf, then copied to c_result,
and the numbers are placed in the array result_buf.
Superstep synchronization is made whenever at each iteration in the main 
routine.
============================================================================*/


#include "bsp.h"
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#define SIZE 32
#define NPROC 8

void print_array(int *array, int n){
  int j;
      printf(" data distribution >>> \n");
      for(j=0;j<n;j++) printf("%d ",array[j]);
      printf("\n");
      fflush(stdout);
}

void copy_array(int *array, int *src_array, int howmany){
     int i;
       for (i=0; i< howmany; i++)
           array[i] = src_array[i];
 }

void  split_array(int *array, int *left_array, int *left_size, int * right_array, int *right_size, int pivot, int all_size){

 int i;
   int  left_ptr = 0;
   int  right_ptr = 0;
   int  ref = array [ pivot];

    for (i=0; i < all_size; i++){
     if (i != pivot) {
        if ( array[i] <  ref)
            left_array[left_ptr++] = *(array+i);
           else right_array[right_ptr++] = *(array+i);
    }
   }
    *left_size = left_ptr;
    *right_size = right_ptr;
}

/* Sequentially doing quick sort of an input array */
void seq_q_sort(int *array, int sz, int start){
   int *pivot = calloc(1, sizeof(int));
   int tmp;
    
  int left_array[SIZE]; /* = calloc(sz, sizeof(int)); */
  int right_array[SIZE]; /* = calloc(sz, sizeof(int)); */
  int *l_size = calloc(1, sizeof(int)); 
  int *r_size = calloc(1, sizeof(int));

  *pivot =  sz/2; 
   tmp = array[*pivot];

   if (sz>=1){
   split_array(array, left_array,l_size, right_array, r_size, *pivot, sz); 
   if(*l_size>= 1)
     seq_q_sort(left_array, *l_size, start);
   if(*r_size>= 1)
     seq_q_sort(right_array, *r_size, start+*l_size);
   copy_array(array, left_array, *l_size);
   array[*l_size] = tmp;
   copy_array(array+*l_size+1, right_array, *r_size);
  }
}

/* Stores job information of the processor to sort the designated data */
void cmd_bsp_put(int active, int who, int level, int start, int old_size, int *work_buf) {
    int work[10];
    work[0]= old_size; /* data size*/
    work[1]= level;  /* level in the job tree starting from level 1 */
    work[2]= start;
/*    work[3]= start;
      work[4]= start; */
    work[5]= active;/* determines if at the level the processor joins sorting*/
    bsp_put(who,work,work_buf,0,10*sizeof(int));/* transfer data to work_buf */
  }

/* Retrieve job data stored by some processor */
void  get_work_data(int *thissize, int *level, int *start, int *work, int *idle) {

  int  tmp [10]; /*= calloc(10, sizeof(int)); */
    bsp_get(bsp_pid(),work, 0, tmp, 10* sizeof(int)) ;

  *thissize = tmp[0];
  *level = tmp[1]; 
  *start = tmp[2]; 
/*   *pivot_index = tmp[3]; */ 
/* *direction = tmp[4]; */ 
  *idle =  tmp[5];
}



void bsp_q_sort(int n, int *w_data, int *buf, int *res_buf, int *count_buf){

  int new_pid;
  int *pivot=calloc(1, sizeof(int)); 
  int *level=calloc(1, sizeof(int)); 
  int *new_pivot= calloc(1, sizeof(int)); 
  int *nct= calloc(1, sizeof(int)); 
  int *thissize = calloc(1, sizeof(int)); 
  int *count=calloc(1, sizeof(int)); 
  int *active = calloc(1, sizeof(int)); 
  int *start = calloc(1, sizeof(int)); 
  int *pivot_data = calloc(1, sizeof(int)); 
  int array[SIZE] ; /* = calloc(n, sizeof(int)); */
  int left_array[SIZE] ; /* = calloc(n, sizeof(int)); */
  int right_array[SIZE] ; /* = calloc(n, sizeof(int)); */
  int *l_size = calloc(1, sizeof(int)); 
  int *r_size = calloc(1, sizeof(int)); 

get_work_data(thissize, level, start, w_data, active);

 if (*active && (*thissize >0)) {

  bsp_get(bsp_pid(), buf, 0, array, *thissize*sizeof(int));
        /* Getting my  data array for sort */

  if ( NPROC/(int) pow(2,*level)<1) {

   seq_q_sort(array, *thissize, *start);
   *nct = *count +*thissize;
   bsp_put(0,thissize,count_buf, bsp_pid()*sizeof(int), sizeof(int)); /*COUNT*/
    printf("\n                                     ===P(%d) SORTED %d NUMBERS *** ===\n", bsp_pid(),*thissize);
    *active =0;  /* TO DEACTIVE MYSELF IF NO JOB BELOW IS GIVEN */
    cmd_bsp_put(*active,bsp_pid(),*level+1,*start,*thissize, w_data);
    bsp_put(0,array, res_buf,*start*sizeof(int),(*thissize)* sizeof(int));/* RECORD */
  }

  else
   {
    new_pid =( bsp_pid() + NPROC/(int)pow(2,*level)) ;
   *pivot = (*thissize + 0)/2; 
   *pivot_data =array[*pivot];
   split_array(array, left_array,l_size, right_array, r_size, *pivot, *thissize); 
   *new_pivot = *start + *l_size ; /* +1 ?*/
   *nct = 1; 
   bsp_put(0, nct, count_buf, bsp_pid()*sizeof(int), sizeof(int));/*COUNT INC*/
    printf("\n                                     ======p(%d) SORTED %d NUMBERS *** ===", bsp_pid(), *nct);

   bsp_put(0,pivot_data,res_buf,*new_pivot*sizeof(int), sizeof(int));/*RECORD*/

       *active = (*l_size >0)? 1:0;   /* LEFT ARRAY */
       bsp_put(bsp_pid(), left_array, buf,0, *l_size*sizeof(int));
       cmd_bsp_put(*active,bsp_pid(),*level+1,*start,*l_size, w_data);


      *active = (*r_size >0)? 1:0;    /* RIGHT ARRAY */
      bsp_put(new_pid, right_array, buf,0, *r_size*sizeof(int));
      cmd_bsp_put(*active,new_pid,*level+1,(*new_pivot)+1,*r_size, w_data);
  }
 }
   else {     ;     }

}



/*** Count the number of values placed in the final position in this iterat.*/
void job_update(int *count_buf){
  int i, sum =0;
  int tmp[NPROC+1] ;
  if (bsp_pid() == 0) {
   bsp_get(0, count_buf,0, tmp, (NPROC+1)*sizeof(int));

       for (i=0; i< NPROC; i++){
       sum += tmp[i];
       tmp[i]=0;
       }
       tmp[NPROC] += sum;
   bsp_put(0, tmp, count_buf,0, (NPROC+1)*sizeof(int));
   bsp_sync();
 }
 else bsp_sync();
}


void main() {
 int i,j,n,last; 
 int count_buf[NPROC+1];
 int c_result[NPROC+1];/* The count of sorted data at each iteration stroed */
 int jobs_done;
 int  xs[SIZE]; /* Array for initial data store */
 int  buf[SIZE]; /* Working buffer to store input, intermediate result */
 int  result[SIZE+1];
 int  work_buf[10]; /* Job command message stored */
 int  res_buf[SIZE+1]; /*  Global memory, stores the sorted output */

bsp_begin(bsp_nprocs());
  n=SIZE;
  
  bsp_pushregister(buf, n* sizeof(int)); 
  bsp_pushregister(work_buf, 10* sizeof(int));
  bsp_pushregister(res_buf, (n+1)*sizeof(int));
  bsp_pushregister(count_buf, (NPROC+1)*sizeof(int));
  bsp_sync();


  for (i=0;i<=NPROC;i++) 
      count_buf[i] =0;
  for (i=0;i<SIZE;i++) 
     res_buf[i] = 0;
  for (i=0;i<n;i++) 
/*    if ( i%2 ) xs[i]=(2*i) % n;
     else xs[i] = (n-i) % n;     */
    if ( i%2 ) xs[i]=(n/2+ i) % n;
     else xs[i] = (i/2) % n;  

  if(bsp_pid() == 0) {printf("    <<< INITIAL "); 
                      print_array(xs,n);
		     }
   if (bsp_pid()==0){
      bsp_put(0, xs, buf, 0, n*sizeof(int));
      cmd_bsp_put(1, 0, 1, 0, n, work_buf); 
    }  
     else {
      cmd_bsp_put(0, bsp_pid(), 1, 0, n, work_buf); 
    }
   i=1;
   jobs_done =0;
   job_update(count_buf);  /* SYNC included in job_update() */
   bsp_sync();
   while ((jobs_done < n) && i<n){

/*  if (bsp_pid() == 0) 
     printf("\n  $$$$$$$    ITER #%d, BY P_%d, JOBS-done=%d\n",
                i, bsp_pid(), jobs_done);   */
   bsp_q_sort(n, work_buf, buf, res_buf, count_buf);
   bsp_sync();

   job_update(count_buf); /* Synch. is included in this subroutine */

   bsp_get(0,count_buf, 0, c_result, (NPROC+1)*sizeof(int)) ;

   bsp_sync();

   jobs_done = c_result[NPROC]; /* The count of the sorted numbers */

   i++;
}

   if(bsp_pid()==0) 
      {printf("    <<< FINAL "); 
                   print_array(res_buf,n); 
      }
   printf(" $$ NOW P[%d] TERMINATING $$\n",  bsp_pid());
   bsp_end();
}


