


























































































































































































#include "bsp.h"
#include "bsp_level1.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>

void bsp_samplesort(int*,int, int **,int *);
void quicksort(int*,int);

char usage_str[] = "main: -v -p <number> -dist <number> -size <number>";
int verbose=0, procs=4, sort_size = 20, sort_dist = 10;


















void bsp_samplesort(int *block, int blocksize, 
                   int **res_block, int *res_blocksize) {
   int nprocs, mypid,offset,i,j,k,l,NKEYS,
       *Samples,*Pivots,
       sample_stride, pivot_stride;

   int left,right,*chunks_size, **chunks;
   int oldtag_size, *temptag,*temppayload;
   int *mergeleft,   mergeleft_size,
       *mergeright,  mergeright_size,
       *mergeres, *mergetmp;

   oldtag_size=sizeof(int);
   bsp_set_tag_size(&oldtag_size);

   nprocs     = bsp_nprocs();
   mypid      = bsp_pid();




   Samples        = calloc(nprocs*nprocs,sizeof(int));
   Pivots         = calloc(nprocs,sizeof(int));
   if (Samples==NULL || Pivots==NULL)
     bsp_abort("{bsp_samplesort} unable to allocate storage");

   bsp_pushregister(block,blocksize*sizeof(int));
   bsp_pushregister(Samples,nprocs*nprocs*sizeof(int));
   bsp_pushregister(Pivots,(nprocs-1)*sizeof(int));
   bsp_sync();




   for(i=0;i<nprocs;i++) 
     bsp_hpput(i,&blocksize,Samples,mypid*sizeof(int),sizeof(int));
   bsp_sync();

   for((i=0,NKEYS=0);i<nprocs;i++) NKEYS+=Samples[i];
   for(i=0;i<nprocs;i++)
     Samples[i] = (int) (( ((float) (Samples[i]*nprocs*nprocs)) / 
                           ((float) NKEYS)) - 0.5);

   for((i=0,j=0);i<nprocs;i++) j+= Samples[i];
   for(i=0;
       (i<nprocs) && (j<(nprocs*nprocs));
       (i++,j++)) Samples[i]++;

#ifdef DEBUG
   for(i=0;i<bsp_nprocs();i++) {
     if (bsp_pid()==i) {
       printf("\nStep 0 [pid %d]: Sample sizes:",bsp_pid());
       for(j=0;j<nprocs;j++) 
         printf("%d ",Samples[j]);
       printf("\n");
     }
     fflush(stdout);
     bsp_sync();
   }
#endif




   sample_stride = (int) ((float)NKEYS / (float)(nprocs*nprocs));
   pivot_stride  = nprocs/2;




   quicksort(block, blocksize);
#ifdef DEBUG
   for(i=0;i<nprocs;i++) {
     if (mypid==i) {
       printf("\nStep 1 [pid %d]: after local sort:",mypid);
       for(j=0;j<blocksize;j++) 
         printf("%d ",block[j]);
       printf("\n");
     }
     fflush(stdout);
     bsp_sync();
   }
#endif




   for((offset=0,i=0);i<mypid;i++) offset+=Samples[i];

   /* Calculate number of samples from me */
   k = Samples[mypid];
   for ((i=0,j=0); j<k; (i+=sample_stride,j++)) {

     bsp_put(0,
             (void*)&block[i],
             (void*)Samples,
             (offset+j)*sizeof(int),
             sizeof(int));
   }
   bsp_sync();   
#ifdef DEBUG
   if (bsp_pid()==0) {
     printf("\nStep 2 [pid %d]: Regular samples:",bsp_pid());
     for(i=0;i<nprocs*nprocs;i++) printf("%d ",Samples[i]);
     printf("\n");
   }
#endif




   if (bsp_pid() == 0) {
     quicksort(Samples, nprocs*nprocs);

#ifdef DEBUG
     printf("\nStep 3 [pid %d]: Sorted samples:",bsp_pid());
     for(i=0;i<nprocs*nprocs;i++) printf("%d ",Samples[i]);
     printf("\n");
#endif
   }




   if (mypid==0) {
     for (i=0; i<nprocs; i++) {
       for(j=1;j<=(nprocs-1);j++) {
         bsp_put(i,
                 (void*)&Samples[(j*nprocs + pivot_stride)-1],
                 (void*)Pivots,
                 (j-1)*sizeof(int),
                 sizeof(int));
         }
      }
   }
   bsp_sync();
#ifdef DEBUG
   for(i=0;i<bsp_nprocs();i++) {
     if (bsp_pid()==i) {
       printf("\nStep 4 [pid %d]: pivots:",bsp_pid());
       for(j=0;j<(nprocs-1);j++) 
         printf("%d ",Pivots[j]);
       printf("\n");
     }
     fflush(stdout);
     bsp_sync();
   }
#endif




   left = 0;
   for(right=0;right<blocksize;right++)
     if (block[right]>=Pivots[0]) break;

   if (right>left)
     bsp_send(0,&mypid,block,(right-left)*sizeof(int));

   left=right;
   for(i=1;i<nprocs-1;i++) {
     while ((block[right]>=Pivots[i-1]) &&
            (block[right]<Pivots[i])) right++;
     if (right>left) 
       bsp_send(i,&mypid,&block[left],(right-left)*sizeof(int));
     left=right;
   }

   if (left<blocksize)
     bsp_send(nprocs-1,&mypid,&block[left],(blocksize-left)*sizeof(int));
   bsp_sync();




   chunks_size = calloc(nprocs,sizeof(int));
   chunks      = calloc(nprocs,sizeof(int*));
   if (chunks_size==NULL || chunks==NULL)
     bsp_abort("Bang");

   for(i=0;i<nprocs;i++) chunks_size[i]=0;
   *res_blocksize=0;
   for(i=0;i<nprocs;i++) {
     j = bsp_hpmove((void**) &temptag,(void**) &temppayload);
     if (j==-1) break;
     else {
       chunks_size[*temptag]= j;
       chunks[*temptag]     = temppayload;
       *res_blocksize      += j / sizeof(int);
     }
   }

   if (*res_blocksize>0) {
     mergeleft = calloc(*res_blocksize,sizeof(int));
     mergeres  = calloc(*res_blocksize,sizeof(int));
   
     if (mergeleft==NULL || mergeres==NULL)
       bsp_abort("{bsp_samplesort} unable to alloc %d elements for result",
                 *res_blocksize);
   }




   mergeleft_size=0;
   for(i=0;i<nprocs;i++)
     if (chunks_size[i]>0) {
       memcpy( mergeleft,
               chunks[i],
               chunks_size[i]);
       mergeleft_size= chunks_size[i]/sizeof(int);  
       break;
   }
   
  
   for(i++;i<nprocs;i++)
     if (chunks_size[i]>0) {
       mergeright_size = chunks_size[i]/sizeof(int);
       mergeright      = chunks[i];
       for((j=0,k=0,l=0);
           (j<mergeleft_size) && (k<mergeright_size);
           l++) {
         if (mergeleft[j]<mergeright[k]) 
           mergeres[l] = mergeleft[j++];
         else
           mergeres[l] = mergeright[k++];
       }
       for(;j<mergeleft_size;j++)  mergeres[l++]=mergeleft[j];
       for(;k<mergeright_size;k++) mergeres[l++]=mergeright[k];

       mergetmp = mergeleft;
       mergeleft = mergeres;
       mergeres  = mergetmp;
       mergeleft_size += mergeright_size;
     }
   if (*res_blocksize != mergeleft_size)
     bsp_abort("{bsp_samplesort} internal error with merging");





   bsp_popregister(Pivots);
   bsp_popregister(Samples);
   bsp_popregister(block);
   free(mergeres);
   free(chunks);
   free(chunks_size);
   free(Pivots);
   free(Samples);
   bsp_set_tag_size(&oldtag_size);
    
   *res_block = mergeleft;
}






void qs(int *array, int left,int right) {
  register int i,j;
  int        x,y;

  i = left; j = right;
  x = array[(left+right)/2];
 
  do {
     while(array[i]<x && i<right) i++;
     while(x<array[j] && j>left)  j--;

     if (i<=j) {
        y = array[i];
        array[i] = array[j];
        array[j] = y;
        i++; j--;
     }
  } while(i<=j);

  if (left<j)  qs(array, left, j);
  if (i<right) qs(array, i, right);
}

void quicksort(int *array, int size) {
  if (size>1) qs(array,0,size-1);
}







static int integer_compare(int *i, int *j)  {
   return(*i - *j);  /* return a +, - or 0 value */
}

void spmd_start() {
  int i,j,chunk_size, new_chunk_size,*array, *result;

  bsp_begin(procs);
    srand(bsp_pid());
    if (bsp_pid()==bsp_nprocs()-1)
      printf("I am the last process (%d) out of %d\n",
             bsp_pid(),bsp_nprocs());
    bsp_pushregister(&sort_size,sizeof(int));
    bsp_pushregister(&verbose,sizeof(int));
    bsp_sync();
    bsp_get(0,&verbose,0,&verbose,sizeof(int));
    bsp_get(0,&sort_size,0,&sort_size,sizeof(int));
    bsp_sync();




    if (bsp_pid() == 0) {
      array = calloc(sort_size,sizeof(int));
      if (array==NULL)
         bsp_abort("Unable to allocate %d elements",sort_size);

      for(i=0;i<sort_size;i++) array[i]=rand() % sort_dist;
      bsp_time();
      qsort(array,sort_size,sizeof(int),
            (int (*)(const void *, const void *)) integer_compare);
      printf("Unix qsort     : %f secs to sort %d elements on 1 proc\n",
             bsp_dtime(),sort_size);
      for(i=0;i<sort_size;i++) array[i]=rand() % sort_dist;
      bsp_time();
      quicksort(array,sort_size);
      printf("C qsort        : %f secs to sort %d elements on 1 proc\n",
              bsp_dtime(),sort_size);
      free(array);
   }
   bsp_sync();





    chunk_size = sort_size / bsp_nprocs();
    if (bsp_pid()==0) chunk_size += sort_size - (chunk_size*bsp_nprocs());
    array = calloc(chunk_size,sizeof(int));
    if (array==NULL) 
      bsp_abort("Unable to allocate %d elements",chunk_size);
    
    for(i=0;i<chunk_size;i++) array[i]=rand() % sort_dist;
   
    if (verbose) {
      for(i=0;i<bsp_nprocs();i++) {
        if (bsp_pid()==i) {
          printf("Before sort on %d array segment of %d elems: ",i,chunk_size);
          for(j=0;j<chunk_size;j++)
            printf("%d, ",array[j]);
          printf("\n");
          fflush(stdout);
        }
        bsp_sync();
      }
    }
    bsp_time();
    bsp_samplesort(array,chunk_size,&result,&new_chunk_size);
    if (bsp_pid()==0)
      printf("BSP sample sort: %f secs to sort %d elements on %d procs\n",
             bsp_dtime(),sort_size,bsp_nprocs());
    if (verbose) {
      for(i=0;i<bsp_nprocs();i++) {
        if (bsp_pid()==i) {
          printf("After sort on %d array segment of %d elems: ",
                 i,new_chunk_size);
          for(j=0;j<new_chunk_size;j++)
            printf("%d, ",result[j]);
          printf("\n");
          fflush(stdout);
        }
        bsp_sync();
      }
    }
     
  bsp_end();
}







void main(int argc,char **argv) {
  int i,temp, command_line_opts=1;
  double trash;

  bsp_init(spmd_start,argc,argv);
  i=1;
  while (command_line_opts) {
    if (i >= argc) {
      command_line_opts = 0;
    } else if ( (strcmp(argv[i],"-size")==0) &&
                (sscanf(argv[i+1],"%d",&temp))) {
      sort_size = temp;
      i += 2;
    } else if ( (strcmp(argv[i],"-dist")==0) &&
                (sscanf(argv[i+1],"%d",&temp))) {
      sort_dist = temp;
      i += 2;
    } else if ( (strcmp(argv[i],"-p")==0) &&
                (sscanf(argv[i+1],"%d",&temp))) {
      procs = temp;
      i += 2;
    } else if (strcmp(argv[i],"-v")==0) {
      verbose =1;
      i += 1;
    } else {
      bsp_abort("{%s}: unknown option \"%s\"\n\t%s\n",
                argv[0],argv[i],usage_str);
    }
  }

  
  spmd_start();
  exit(0);
}










































































